/* Author:  Robbert van Renesse, Cornell University.
 *
 * This library maintains value_handles.  A value handle typically refers to some
 * Ocaml value.  The value itself is kept in a tuple, which in turn is registered
 * as a global root.  The handle on the value is the index into this tuple.  The
 * addresses of all handles are known to this module, and can be checked for
 * validity.  This is necessary, since several routines accept either a handle or
 * something else, and this is the only way in which they can be distinguished.
 */

#include "memory.h"
#include "camou.h"

static value camou_values;	/* This points to the tuple with the actual values */
static value camou_nvals;	/* Size of this tuple */

static value_handle camou_free;	/* Free list of these handles (contains at least one) */
static value_handle camou_used;	/* List of the handles that are in use */

/* Allocate a value handle, and fill it with the given value.  If there are
 * none left, increase the size of the values tuple.
 */
value_handle camou_alloc(value v){
	extern char *malloc();
	value_handle vh;
	int i, newsize;

	vh = camou_free;
	camou_free = vh->next;
	if ((vh->next = camou_used) != 0)
		camou_used->prev = &vh->next;
	vh->prev = &camou_used;
	camou_used = vh;
	modify(&Field(camou_values, vh->index), v);
	if (camou_free == 0) {
		newsize = camou_nvals * 2;
		v = alloc_shr(newsize, 0);
		for (i = 0; i < camou_nvals; i++)
			initialize(&Field(v, i), Field(camou_values, i));
		for (; i < newsize; i++) {
			initialize(&Field(v, i), Val_unit);
			vh = (value_handle) malloc(sizeof(*vh));
			vh->index = i;
			vh->next = camou_free;
			camou_free = vh;
		}
		modify(&camou_values, v);
		camou_nvals = newsize;
	}
	return camou_used;
}

/* See if this is one of my handles.  If so, return the corresponding value.
 */
int camou_ishandle(void *vh, value *vp){
	value_handle u;

	for (u = camou_used; u != 0; u = u->next)
		if (u == vh) {
			if (vp != 0)
				*vp = Field(camou_values, u->index);
			return 1;
		}
	return 0;
}

/* Release this handle.
 */
void camou_release(value_handle vh){
	if ((*vh->prev = vh->next) != 0)
		vh->next->prev = vh->prev;
	vh->next = camou_free;
	camou_free = vh;
}

/* Initialize this module.
 */
value camou_init(value unit){
	register_global_root(&camou_values);
	camou_values = alloc_tuple(1);
	Field(camou_values, 0) = Val_unit;
	camou_nvals = 1;
	camou_free = (value_handle) malloc(sizeof(*camou_free));
	camou_free->index = 0;
	camou_free->next = 0;
	return Val_unit;
}

/*******************************************************************/

value c2ml_string(string s){
	value v;

	if (camou_ishandle((value_handle) s, &v))
		return v;
	else
		return copy_string(s);
}

/* If upcall is a handle, return the corresponding value.
 * Otherwise generate a closure value by partially evaluating
 * cb with an upcall argument.
 */
value c2ml_callback(value cb, void *upcall){
	value v;

	if (camou_ishandle(upcall, &v))
		return v;
	else
		return callback(cb, (value) upcall);
}

value c2ml_abstract(abstract_t a){
	return Field(camou_values, ((value_handle) a)->index);
}

value c2ml_alpha(alpha a){
	return Field(camou_values, ((value_handle) a)->index);
}

string ml2c_string(enum camou_direction dir, value v){
	if (dir == CAMOU_ARG)
		return String_val(v);
	else
		return (string) camou_alloc(v);
}

void *ml2c_upcall(enum camou_direction dir, value v){
	if (dir == CAMOU_ARG)
		switch (Tag_val(v)) {
		case Closure_tag:
			return camou_alloc(v);
		case Closure_tag + 1:
			return (void *) Field(v, 0);
		default:
			failwith("ml2c_mlc: bad tag");
		}
	else
		return (void *) camou_alloc(v);
}
